#include "types.h"
#include "riscv.h"
#include "defs.h"
#include "param.h"
#include "memlayout.h"
#include "spinlock.h"

#define NSHARED 16
#define MAXSHAREDPAGES 64

struct shared_mem {
    int key;
    int ref_count;
    void *pages[MAXSHAREDPAGES];
    int size;
};

struct shared_mem shared_mem_segments[NSHARED];
struct spinlock shm_lock;

void
shm_init(void)
{
    initlock(&shm_lock, "shm_lock");
    for (int i = 0; i < NSHARED; i++) {
        shared_mem_segments[i].key = -1;
        shared_mem_segments[i].ref_count = 0;
        shared_mem_segments[i].size = 0;
        for (int j = 0; j < MAXSHAREDPAGES; j++) {
            shared_mem_segments[i].pages[j] = 0;
        }
    }
}

uint64
shm_alloc(int key, int size)
{
    acquire(&shm_lock);

    // 检查是否已经存在相同 key 的共享内存段
    for (int i = 0; i < NSHARED; i++) {
        if (shared_mem_segments[i].key == key) {
            shared_mem_segments[i].ref_count++;
            release(&shm_lock);
            return key;
        }
    }

    // 分配新的共享内存段
    for (int i = 0; i < NSHARED; i++) {
        if (shared_mem_segments[i].key == -1) {
            shared_mem_segments[i].key = key;
            shared_mem_segments[i].ref_count = 1;
            shared_mem_segments[i].size = size;

            // 为共享内存分配物理页
            for (int j = 0; j < size; j++) {
                shared_mem_segments[i].pages[j] = kalloc();
                if (!shared_mem_segments[i].pages[j]) {
                    release(&shm_lock);
                    return -1;
                }
                memset(shared_mem_segments[i].pages[j], 0, PGSIZE);
            }

            release(&shm_lock);
            return key;
        }
    }

    release(&shm_lock);
    return -1; // 共享内存分配失败
}

uint64
shm_map(int key, uint64 addr)
{
    acquire(&shm_lock);

    // 查找 key 对应的共享内存段
    struct shared_mem *shm = 0;
    for (int i = 0; i < NSHARED; i++) {
        if (shared_mem_segments[i].key == key) {
            shm = &shared_mem_segments[i];
            break;
        }
    }

    if (!shm) {
        release(&shm_lock);
        return -1; // 找不到共享内存段
    }

    struct proc *p = myproc();
    for (int i = 0; i < shm->size; i++) {
        if (mappages(p->pagetable, addr + i * PGSIZE, PGSIZE, (uint64)shm->pages[i], PTE_W | PTE_R | PTE_U) < 0) {
            release(&shm_lock);
            return -1; // 映射失败
        }
    }

    release(&shm_lock);
    return addr; // 映射成功
}
